from logging import getLogger

from .autodiscover import discover
from .configuration import Configuration
from .credentials import Credentials, DELEGATE, IMPERSONATION
from .errors import ErrorFolderNotFound, ErrorAccessDenied
from .folders import Root, Calendar, Inbox, Tasks, Contacts, SHALLOW, DEEP
from .protocol import Protocol
from .ewsdatetime import EWSDateTime, EWSTimeZone

log = getLogger(__name__)


class Account:
    """
    Models an Exchange server user account. The primary key for accounts is their PrimarySMTPAddress
    """
    def __init__(self, primary_smtp_address, fullname=None, credentials=None, access_type=None, autodiscover=False,
                 config=None):
        if '@' not in primary_smtp_address:
            raise ValueError("primary_smtp_address '%s' is not an email address" % primary_smtp_address)
        self.primary_smtp_address = primary_smtp_address
        self.fullname = fullname
        if not (credentials or config):
            raise AttributeError('Either config or credentials must be supplied')
        self.credentials = credentials or config.credentials
        # Assume delegate access if individual credentials are provided. Else, assume service user with impersonation
        self.access_type = access_type or (DELEGATE if credentials else IMPERSONATION)
        assert self.access_type in (DELEGATE, IMPERSONATION)
        if autodiscover:
            self.primary_smtp_address, self.protocol = discover(email=self.primary_smtp_address,
                                                                credentials=self.credentials)
            if config:
                assert isinstance(config, Configuration)
        else:
            self.protocol = config.protocol
        # We may need to override the default server version on a per-account basis because Microsoft may report one
        # server version up-front but delegate account requests to an older backend server.
        self.version = self.protocol.version
        self.root = Root(self)
        self.folders = {}  # TODO Unimplemented - should support Inbox, Tasks, Contacts
        # If the account contains a shared calendar from a different user, that calendar will be in the folder list.
        # Attempt not to return one of those. An account may not always have a calendar called "Calendar", but a
        # Calendar folder with a localized name instead. Return that, if it's available.
        try:
            # Get the default calendar
            self.calendar = Calendar(self).get_folder()
        except ErrorAccessDenied:
            self.calendar = Calendar(self)
            dt = EWSTimeZone.timezone('UTC').localize(EWSDateTime(2000, 1, 1))
            self.calendar.find_items(start=dt, end=dt, categories=['DUMMY'])
            # Maybe we just don't have GetFolder access
        except ErrorFolderNotFound as e:
            # Try to guess which calendar folder is the default
            folders = self.root.get_folders(depth=SHALLOW)
            # This is a folder available in some Exchange accounts. It only contains folders owned by the account.
            has_tois = False
            for folder in folders:
                if folder.name == 'Top of Information Store':
                    has_tois = True
                    folders = folder.get_folders(depth=SHALLOW)
                    break
            if not has_tois:
                folders = self.root.get_folders(depth=DEEP)
            calendars = []
            for folder in folders:
                # Use the calendar wth a localized name
                if folder.folder_class != folder.CONTAINER_CLASS:
                    # This is a pseudo-folder
                    continue
                if folder.name.title() in Calendar.LOCALIZED_NAMES:
                    calendars.append(folder)
            if not calendars:
                # Use the distinguished folder
                for folder in folders:
                    if folder.folder_class != folder.CONTAINER_CLASS and folder.is_distinguished:
                        calendars.append(folder)
            if not calendars:
                raise ErrorFolderNotFound('No useable calendar folders') from e
            assert len(calendars) == 1, 'Multiple calendars could be default: %s' % [str(f) for f in calendars]
            self.calendar = calendars[0]

        assert isinstance(self.credentials, Credentials)
        assert isinstance(self.protocol, Protocol)
        log.debug('Added account: %s', self)

    @property
    def inbox(self):
        folder = None
        for folder in self.folders[Inbox]:
            if folder.is_distinguished:
                return folder
        return folder

    @property
    def tasks(self):
        folder = None
        for folder in self.folders[Tasks]:
            if folder.is_distinguished:
                return folder
        return folder

    @property
    def contacts(self):
        folder = None
        for folder in self.folders[Contacts]:
            if folder.is_distinguished:
                return folder
        return folder

    def get_domain(self):
        return self.primary_smtp_address.split('@')[1].lower().strip()

    def __str__(self):
        txt = '%s' % self.primary_smtp_address
        if self.fullname:
            txt += ' (%s)' % self.fullname
        return txt



#!/usr/bin/env python3
import os
from setuptools import setup


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()

setup(
    name='exchangelib',
    version='1.3.2',
    author='Erik Cederstrand',
    author_email='erik@cederstrand.dk',
    description='Client for Microsoft Exchange Web Services (EWS)',
    long_description=read('README.rst'),
    license='BSD',
    keywords='Exchange EWS autodiscover',
    install_requires=['requests>=2.7', 'requests-ntlm>=0.2.0', 'dnspython3>=1.12.0', 'pytz', 'lxml'],
    packages=['exchangelib'],
    tests_require=['PyYAML'],
    test_suite='tests',
    zip_safe=False,
    url='https://github.com/ecederstrand/exchangelib',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Communications',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3 :: Only',
    ],
)



import logging
from xml.etree.cElementTree import ParseError

import requests.sessions
import requests.adapters

from .errors import UnauthorizedError, TransportError, EWSWarning
from .transport import TNS, SOAPNS, dummy_xml, get_auth_instance
from .util import is_xml, to_xml, post_ratelimited

log = logging.getLogger(__name__)

# Legend for dict:
#   Key: shortname
#   Values: (EWS API version ID, full name)

# 'shortname' comes from types.xsd and is the official version of the server, corresponding to the version numbers
# supplied in SOAP headers. 'API version' is the version name supplied in the RequestServerVersion element in SOAP
# headers and describes the EWS API version the server accepts. Valid values for this element are described here:
#    http://msdn.microsoft.com/en-us/library/bb891876(v=exchg.150).aspx

VERSIONS = {
    'Exchange2007': ('Exchange2007', 'Microsoft Exchange Server 2007'),
    'Exchange2007_SP1': ('Exchange2007_SP1', 'Microsoft Exchange Server 2007 SP1'),
    'Exchange2007_SP2': ('Exchange2007_SP1', 'Microsoft Exchange Server 2007 SP2'),
    'Exchange2007_SP3': ('Exchange2007_SP1', 'Microsoft Exchange Server 2007 SP3'),
    'Exchange2010': ('Exchange2010', 'Microsoft Exchange Server 2010'),
    'Exchange2010_SP1': ('Exchange2010_SP1', 'Microsoft Exchange Server 2010 SP1'),
    'Exchange2010_SP2': ('Exchange2010_SP2', 'Microsoft Exchange Server 2010 SP2'),
    'Exchange2010_SP3': ('Exchange2010_SP2', 'Microsoft Exchange Server 2010 SP3'),
    'Exchange2013': ('Exchange2013', 'Microsoft Exchange Server 2013'),
    'Exchange2013_SP1': ('Exchange2013_SP1', 'Microsoft Exchange Server 2013 SP1'),
    'Exchange2015': ('Exchange2015', 'Microsoft Exchange Server 2015'),
    'Exchange2015_SP1': ('Exchange2015_SP1', 'Microsoft Exchange Server 2015 SP1'),
    'Exchange2016': ('Exchange2016', 'Microsoft Exchange Server 2016'),
}

# List of build numbers here: https://technet.microsoft.com/en-gb/library/hh135098(v=exchg.150).aspx
API_VERSION_FROM_BUILD_NUMBER = {
    8: {
        0: 'Exchange2007',
        1: 'Exchange2007_SP1',
        2: 'Exchange2007_SP1',
        3: 'Exchange2007_SP1',
    },
    14: {
        0: 'Exchange2010',
        1: 'Exchange2010_SP1',
        2: 'Exchange2010_SP2',
        3: 'Exchange2010_SP2',
    },
    15: {
        0: 'Exchange2013',  # Minor builds above 847 are Exchange2013_SP1
        1: 'Exchange2016',
    },
}

# Build a list of unique API versions, used when guessing API version supported by the server.  Use reverse order so we
# get the newest API version supported by the server.
API_VERSIONS = sorted({v[0] for v in VERSIONS.values()}, reverse=True)


class Version:
    """
    Holds information about the server version
    """

    def __init__(self, major_version, minor_version, major_build, minor_build, api_version):
        self.major_version = major_version
        self.minor_version = minor_version
        self.major_build = major_build
        self.minor_build = minor_build
        self.api_version = api_version
        if major_version < 8:
            raise ValueError("Exchange major versions below 8 don't support EWS (%s)", str(self))

    @property
    def build(self):
        return '%s.%s.%s.%s' % (self.major_version, self.minor_version, self.major_build, self.minor_build)

    @property
    def fullname(self):
        return VERSIONS[self.api_version][1]

    @classmethod
    def guess(cls, protocol):
        """
        Tries to ask the server which version it has. We haven't set up an Account object yet, so generate a request
        by hand. We only need a response header containing a ServerVersionInfo element. Apparently, EWS has no problem
        supplying one version in its types.xsd and reporting another in its SOAP headers. Trust the SOAP version.
        """
        log.debug('Asking server for version info')
        # Can't use a session object from the protocol pool for docs because sessions are created with service auth.
        try:
            auth = get_auth_instance(credentials=protocol.credentials, auth_type=protocol.docs_auth_type)
            shortname = cls._get_shortname_from_docs(auth=auth, types_url=protocol.types_url)
            log.debug('Shortname according to %s: %s', protocol.types_url, shortname)
        except (TransportError, UnauthorizedError) as e:
            log.warning(str(e))
            shortname = None
        api_version = VERSIONS[shortname][0] if shortname else None
        return cls._guess_version_from_service(protocol=protocol, ews_url=protocol.ews_url, hint=api_version)

    @classmethod
    def _get_shortname_from_docs(cls, auth, types_url):
        # Get the server version from types.xsd. A server response provides the build numbers. We can't necessarily use
        # the service auth type since it may not be the same as the auth type for docs.
        log.debug('Getting %s with auth type %s', types_url, auth.__class__.__name__)
        # Some servers send an empty response if we send 'Connection': 'close' header
        with requests.sessions.Session() as s:
            r = s.get(url=types_url, auth=auth, allow_redirects=False, stream=False)
        log.debug('Request headers: %s', r.request.headers)
        log.debug('Response code: %s', r.status_code)
        log.debug('Response headers: %s', r.headers)
        if r.status_code == 401:
            raise UnauthorizedError('Wrong username or password for %s' % types_url)
        if r.status_code == 302:
            log.debug('We were redirected. Cant get version info from docs')
            return None
        if r.status_code == 503:
            log.debug('Service is unavailable. Cant get version info from docs')
            return None
        if r.status_code != 200:
            if 'The referenced account is currently locked out' in r.text:
                raise TransportError('The service account is currently locked out')
            raise TransportError('Unexpected HTTP status %s when getting %s (%s)' % (r.status_code, types_url, r.text))
        if not is_xml(r.text):
            raise TransportError('Unexpected result when getting %s. Maybe this is not an EWS server?%s' % (
                types_url,
                '\n\n%s[...]' % r.text[:200] if len(r.text) > 200 else '\n\n%s' % r.text if r.text else '',
            ))
        return to_xml(r.text, encoding=r.encoding).get('version')

    @classmethod
    def _guess_version_from_service(cls, protocol, ews_url, hint=None):
        # We need to guess the version. If we got a shortname from docs, start guessing that
        if hint:
            api_versions = [hint] + [v for v in API_VERSIONS if v != hint]
        else:
            api_versions = API_VERSIONS
        for api_version in api_versions:
            try:
                return cls._get_version_from_service(protocol=protocol, ews_url=ews_url, api_version=api_version)
            except EWSWarning:
                continue
        raise TransportError('Unable to guess version')

    @classmethod
    def _get_version_from_service(cls, protocol, ews_url, api_version):
        assert api_version
        xml = dummy_xml(version=api_version)
        # Create a minimal, valid EWS request to force Exchange into accepting the request and returning EWS xml
        # containing server version info. Some servers will only reply with their version if a valid POST is sent.
        session = protocol.get_session()
        log.debug('Test if service API version is %s using auth %s', api_version, session.auth.__class__.__name__)
        r, session = post_ratelimited(protocol=protocol, session=session, url=ews_url, headers=None, data=xml,
                                      timeout=protocol.timeout, verify=True, allow_redirects=False)
        protocol.release_session(session)

        if r.status_code == 401:
            raise UnauthorizedError('Wrong username or password for %s' % ews_url)
        elif r.status_code == 302:
            log.debug('We were redirected. Cant get version info from docs')
            return None
        elif r.status_code == 503:
            log.debug('Service is unavailable. Cant get version info from docs')
            return None
        if r.status_code == 400:
            raise EWSWarning('Bad request')
        if r.status_code == 500 and ('The specified server version is invalid' in r.text or
                                     'ErrorInvalidSchemaVersionForMailboxVersion' in r.text):
            raise EWSWarning('Invalid server version')
        if r.status_code != 200:
            if 'The referenced account is currently locked out' in r.text:
                raise TransportError('The service account is currently locked out')
            raise TransportError('Unexpected HTTP status %s when getting %s (%s)' % (r.status_code, ews_url, r.text))
        log.debug('Response data: %s', r.text)
        try:
            header = to_xml(r.text, encoding=r.encoding).find('{%s}Header' % SOAPNS)
            if not header:
                raise ParseError()
        except ParseError as e:
            raise EWSWarning('Unknown XML response from %s (response: %s)' % (ews_url, r.text)) from e
        info = header.find('{%s}ServerVersionInfo' % TNS)
        if info is None:
            raise TransportError('No ServerVersionInfo in response: %s' % r.text)

        version = cls.from_response(requested_api_version=api_version, response=r)
        log.debug('Service version is: %s', version)
        return version

    @classmethod
    def from_response(cls, requested_api_version, response):
        try:
            header = to_xml(response.text, encoding=response.encoding).find('{%s}Header' % SOAPNS)
            if not header:
                raise ParseError()
        except ParseError as e:
            raise EWSWarning('Unknown XML response from %s (response: %s)' % (response, response.text)) from e
        info = header.find('{%s}ServerVersionInfo' % TNS)
        if info is None:
            raise TransportError('No ServerVersionInfo in response: %s' % response.text)

        major_version, minor_version, major_build, minor_build = \
            [int(info.get(k)) for k in ['MajorVersion', 'MinorVersion', 'MajorBuildNumber', 'MinorBuildNumber']]
        for k, v in dict(MajorVersion=major_version, MinorVersion=minor_version, MajorBuildNumber=major_build,
                         MinorBuildNumber=minor_build).items():
            if v is None:
                raise TransportError('No %s in response: %s' % (k, response.text))
        api_version_from_server = info.get('Version')
        if api_version_from_server is None:
            # Not all Exchange servers send the Version element
            api_version_from_server = cls.api_version_from_build_number(major_version, minor_version, major_build)
        if api_version_from_server != requested_api_version:
            if api_version_from_server.startswith('V2_') \
                    or api_version_from_server.startswith('V2015_') \
                    or api_version_from_server.startswith('V2016_'):
                # Office 365 is an expert in sending invalid server versions...
                log.info('API version "%s" worked but server reports version "%s". Using "%s"', requested_api_version,
                         api_version_from_server, requested_api_version)
                api_version_from_server = requested_api_version
            else:
                # Work around a bug in Exchange that reports a bogus API version in the XML response. Trust server
                # response except 'V2_nn' or 'V201[5,6]_nn_mm' which is bogus
                log.info('API version "%s" worked but server reports version "%s". Using "%s"', requested_api_version,
                         api_version_from_server, api_version_from_server)
        return cls(major_version, minor_version, major_build, minor_build, api_version_from_server)

    @staticmethod
    def api_version_from_build_number(major_version, minor_version, major_build):
        api_version = API_VERSION_FROM_BUILD_NUMBER[major_version][minor_version]
        if major_version == 15 and major_version == 0 and major_build >= 847:
            api_version = 'Exchange2013_SP1'
        return api_version

    def __str__(self):
        return 'Build=%s, API=%s, Fullname=%s' % (self.build, self.api_version, self.fullname)



"""
Implements an Exchange user object and access types. Exchange provides two different ways of granting access for a
login to a specific account. Impersonation is used mainly for service accounts that connect via EWS. Delegate is used
for ad-hoc access e.g. granted manually by the user.
See http://blogs.msdn.com/b/exchangedev/archive/2009/06/15/exchange-impersonation-vs-delegate-access.aspx
"""
import logging

log = logging.getLogger(__name__)

IMPERSONATION = 'impersonation'
DELEGATE = 'delegate'

EMAIL = 'email'
DOMAIN = 'domain'
UPN = 'upn'


class Credentials:
    # Keeps login info the way Exchange likes it. Usernames for authentication are of one of these forms:
    #
    #    * PrimarySMTPAddress
    #    * WINDOMAIN\username
    #    * User Principal Name (UPN)
    def __init__(self, username, password):
        if username.count('@') == 1:
            self.type = EMAIL
        elif username.count('\\') == 1:
            self.type = DOMAIN
        else:
            self.type = DOMAIN
        self.username = username
        self.password = password

    def __str__(self):
        return self.username



import logging
from xml.etree.cElementTree import Element, tostring

import requests.sessions
from requests.auth import HTTPBasicAuth, HTTPDigestAuth
from requests_ntlm import HttpNtlmAuth

from .credentials import IMPERSONATION, EMAIL
from .errors import UnauthorizedError, TransportError, RedirectError
from .util import set_xml_attr, is_xml, get_redirect_url

log = logging.getLogger(__name__)

# XML namespaces
SOAPNS = 'http://schemas.xmlsoap.org/soap/envelope/'
MNS = 'http://schemas.microsoft.com/exchange/services/2006/messages'
TNS = 'http://schemas.microsoft.com/exchange/services/2006/types'
ENS = 'http://schemas.microsoft.com/exchange/services/2006/errors'

# Authentication method enums
NOAUTH = 'no authentication'
NTLM = 'NTLM'
BASIC = 'basic'
DIGEST = 'digest'
UNKNOWN = 'unknown'

AUTH_TYPE_MAP = {
    NTLM: HttpNtlmAuth,
    BASIC: HTTPBasicAuth,
    DIGEST: HTTPDigestAuth,
    NOAUTH: None,
}

AUTH_CLASS_MAP = dict((v, k) for k, v in AUTH_TYPE_MAP.items())


def test_credentials(protocol):
    return _test_docs_credentials(protocol) and _test_service_credentials(protocol)


def _test_docs_credentials(protocol):
    log.debug("Trying auth type '%s' on '%s'", protocol.docs_auth_type)
    # Retrieve the result. We allow 401 errors to happen since the authentication type may be wrong, giving a 401
    # response.
    auth = get_auth_instance(credentials=protocol.credentials, auth_type=protocol.docs_auth_type)
    with requests.sessions.Session() as s:
        r = s.get(url=protocol.types_url, auth=auth, allow_redirects=False)
    return _test_response(auth=auth, response=r)


def _test_service_credentials(protocol):
    log.debug("Trying auth type '%s' on '%s'", protocol.ews_auth_type, protocol.ews_url)
    # Retrieve the result. We allow 401 errors to happen since the authentication type may be wrong, giving a 401
    # response.
    headers = {'Content-Type': 'text/xml; charset=utf-8'}
    data = dummy_xml(version=protocol.version.api_version)
    auth = get_auth_instance(credentials=protocol.credentials, auth_type=protocol.ews_auth_type)
    with requests.sessions.Session() as s:
        r = s.post(url=protocol.ews_url, headers=headers, data=data, auth=auth, allow_redirects=False)
    return _test_response(auth=auth, response=r)


def _test_response(auth, response):
    log.debug('Response headers: %s', response.headers)
    resp = response.text
    log.debug('Response data: %s [...]', str(resp[:1000]))
    if is_xml(resp):
        log.debug('This is XML')
        # Assume that any XML response is good news
        return True
    elif _is_unauthorized(resp):
        # Exchange brilliantly sends an unauth message as a non-401 page. Clever.
        raise UnauthorizedError('Unauthorized (non-401)')
    elif isinstance(auth, HttpNtlmAuth) and not resp:
        # It seems the NTLM handler doesn't throw 401 errors. If the request is invalid, it doesn't bother
        # responding with anything. Even more clever.
        raise UnauthorizedError('Unauthorized (NTLM, empty response)')
    else:
        raise TransportError('Unknown response from Exchange:\n\n%s' % resp)


def _is_unauthorized(txt):
    """
    Helper function. Test if response contains an "Unauthorized" message
    """
    if txt.lower().count('unauthorized') > 0:
        return True
    return False


def wrap(content, version, account, ewstimezone=None, encoding='utf-8'):
    """
    Generate the necessary boilerplate XML for a raw SOAP request. The XML is specific to the server version.
    ExchangeImpersonation allows to act as the user we want to impersonate.
    """
    envelope = Element('s:Envelope')
    envelope.set('xmlns:s', SOAPNS)
    envelope.set('xmlns:t', TNS)
    envelope.set('xmlns:m', MNS)
    header = Element('s:Header')
    requestserverversion = Element('t:RequestServerVersion', Version=version)
    header.append(requestserverversion)
    if account and account.access_type == IMPERSONATION:
        exchangeimpersonation = Element('t:ExchangeImpersonation')
        connectingsid = Element('t:ConnectingSID')
        set_xml_attr(connectingsid, 't:PrimarySmtpAddress', account.primary_smtp_address)
        exchangeimpersonation.append(connectingsid)
        header.append(exchangeimpersonation)
    if ewstimezone:
        timezonecontext = Element('t:TimeZoneContext')
        timezonedefinition = Element('t:TimeZoneDefinition', Id=ewstimezone.ms_id)
        timezonecontext.append(timezonedefinition)
        header.append(timezonecontext)
    envelope.append(header)
    body = Element('s:Body')
    body.append(content)
    envelope.append(body)
    return (b'<?xml version="1.0" encoding="' + encoding.encode(encoding)) + b'"?>' \
        + tostring(envelope, encoding=encoding)


def get_auth_instance(credentials, auth_type):
    """
    Returns an *Auth instance suitable for the requests package
    """
    try:
        model = AUTH_TYPE_MAP[auth_type]
    except KeyError as e:
        raise ValueError("Authentication type '%s' not supported" % auth_type) from e
    else:
        if model is None:
            return None
        username = credentials.username
        if auth_type == NTLM and credentials.type == EMAIL:
            username = '\\' + username
        return model(username=username, password=credentials.password)


def get_auth_type(auth):
    try:
        return AUTH_CLASS_MAP[auth.__class__]
    except KeyError as e:
        raise ValueError("Authentication model '%s' not supported" % auth.__class__) from e


def get_autodiscover_authtype(server, has_ssl, url, data, timeout):
    # First issue a HEAD request to look for a location header. This is the autodiscover HTTP redirect method. If there
    # was no redirect, continue trying a POST request with a valid payload.
    log.debug('Getting autodiscover auth type for %s %s', url, timeout)
    headers = {'Content-Type': 'text/xml; charset=utf-8'}
    with requests.sessions.Session() as s:
        r = s.head(url=url, headers=headers, timeout=timeout, allow_redirects=False, verify=True)
        if r.status_code == 302:
            redirect_url, redirect_server, redirect_has_ssl = get_redirect_url(r, server, has_ssl)
            log.debug('Autodiscover HTTP redirect to %s', redirect_url)
            if not (server == redirect_server and has_ssl == redirect_has_ssl):
                raise RedirectError(url=redirect_url, server=redirect_server, has_ssl=redirect_has_ssl)
            # Some MS servers are masters of fucking up HTTP, issuing 302 to an error page with zero content. Give this
            # URL a chance with a POST request.
            # raise TransportError('Circular redirect')
        r = s.post(url=url, headers=headers, data=data, timeout=timeout, allow_redirects=False, verify=True)
    return _get_auth_method_from_response(server=server, response=r, has_ssl=has_ssl)


def get_docs_authtype(server, has_ssl, url):
    # Get auth type by tasting headers from the server. Don't do HEAD requests. It's too error prone.
    log.debug('Getting docs auth type for %s', url)
    headers = {'Content-Type': 'text/xml; charset=utf-8'}
    with requests.sessions.Session() as s:
        r = s.get(url=url, headers=headers, allow_redirects=True, verify=True)
    return _get_auth_method_from_response(server=server, response=r, has_ssl=has_ssl)


def get_service_authtype(server, has_ssl, ews_url, versions):
    # Get auth type by tasting headers from the server. Only do post requests. HEAD is too error prone, and some servers
    # are set up to redirect to OWA on all requests except POST to /EWS/Exchange.asmx
    log.debug('Getting service auth type for %s', ews_url)
    headers = {'Content-Type': 'text/xml; charset=utf-8'}
    # TODO We don't know the version yet, but we need it to create a valid request because some Exchange servers
    # only respond when given a valid request. Try all known versions. Gross.
    with requests.sessions.Session() as s:
        for version in versions:
            data = dummy_xml(version=version)
            log.debug('Requesting %s from %s', data, ews_url)
            r = s.post(url=ews_url, headers=headers, data=data, allow_redirects=True, verify=True)
            auth_method = _get_auth_method_from_response(server=server, response=r, has_ssl=has_ssl)
            if auth_method != UNKNOWN:
                return auth_method
            raise ValueError("Authentication type '%s' not supported" % auth_method)


def _get_auth_method_from_response(server, response, has_ssl):
    # First, get the auth method from headers. Then, test credentials. Don't handle redirects - burden is on caller.
    log.debug('Request headers: %s', response.request.headers)
    log.debug('Response headers: %s', response.headers)
    if response.status_code == 200:
        log.debug('No authentication needed')
        return NOAUTH
    if response.status_code == 302:
        # Some servers are set up to redirect to OWA on all requests except POST to EWS/Exchange.asmx
        redirect_url, redirect_server, redirect_has_ssl = get_redirect_url(response, server, has_ssl)
        if server == redirect_server and has_ssl == redirect_has_ssl:
            raise TransportError('Circular redirect')
        raise RedirectError(url=redirect_url, server=redirect_server, has_ssl=redirect_has_ssl)
    if response.status_code != 401:
        raise TransportError('Unexpected response: %s %s' % (response.status_code, response.reason))

    # Get auth type from headers
    for key, val in response.headers.items():
        if key.lower() == 'www-authenticate':
            vals = _tokenize(val.lower())
            for v in vals:
                if v.startswith('realm'):
                    realm = v.split('=')[1].strip('"')
                    log.debug('realm: %s', realm)
            # Prefer most secure auth method if more than one is offered. See discussion at
            # http://docs.oracle.com/javase/7/docs/technotes/guides/net/http-auth.html
            if 'digest' in vals:
                log.debug('Auth type is %s', DIGEST)
                return DIGEST
            if 'ntlm' in vals:
                log.debug('Auth type is %s', NTLM)
                return NTLM
            if 'basic' in vals:
                log.debug('Auth type is %s', BASIC)
                return BASIC
    raise UnauthorizedError('Got a 401, but no compatible auth type was reported by server')


def _tokenize(val):
    # Splits cookie auth values
    tokens = []
    token = ''
    quote = False
    for c in val:
        if c in (' ', ',') and not quote:
            if token not in ('', ','):
                tokens.append(token)
            token = ''
            continue
        elif c == '"':
            token += c
            if quote:
                tokens.append(token)
                token = ''
            quote = not quote
            continue
        token += c
    if token:
        tokens.append(token)
    return tokens


def dummy_xml(version):
    # Used as a minimal, valid EWS request to force Exchange into accepting the request and returning EWS XML
    # containing server version info.
    from .services import ResolveNames  # Avoid circular import
    return ResolveNames(protocol=None).payload(version=version, account=None, unresolvedentries=['DUMMY'])
